from dataclasses import dataclass
from typing import Dict, Sequence, Optional, List
from sqlalchemy import select, update, delete


from ._model import Model
from ._connection import make_session


@dataclass
class Entity:
    """Entity model."""

    model: Model

    # TODO: filter
    # TODO: ordering
    # TODO: pagination params

    i18n = 'ru'
    i18n_fields: Sequence[str] = None

    @classmethod
    async def read(cls, id: int, lang: str = None) -> Optional[Dict]:
        """Read entity or return None."""
        async with make_session() as session:
            raw = await cls.get_by_id(session, id)
            cls.i18n_release(raw, lang=lang)
            return raw

    @classmethod
    async def read_all(cls, lang: str = None) -> List[Dict]:
        """Read all entities.
        TODO: add pagination only here.
        """
        results = []
        async with make_session() as session:
            query = await session.execute(select(cls.model))
            rows = query.scalars()
            if not rows:
                return results

            for row in list(rows):
                raw = cls.row_to_dict(row)
                cls.i18n_release(raw, lang=lang)
                results.append(raw)

            return results

    @classmethod
    def i18n_prepare(cls, fields: Dict) -> None:
        """Provide i18n support."""
        if cls.i18n_fields:
            for i18n_field in cls.i18n_fields:
                _rec = {cls.i18n: str(fields.get(i18n_field))}
                if i18n_field in fields:
                    fields[i18n_field] = _rec

    @classmethod
    def i18n_release(cls, fields: Dict, lang: str = None):
        """Release i18n fields."""
        if cls.i18n_fields:
            for i18n_field in cls.i18n_fields:
                if i18n_field in fields:
                    _rec = fields[i18n_field].get(cls.i18n, '')
                    if lang and fields[i18n_field].get(lang):
                        _rec = fields[i18n_field].get(lang)
                    fields[i18n_field] = _rec
        return fields

    @classmethod
    async def create(cls, fields: Dict) -> Dict:
        """Create entity from dict and save into database."""
        cls.i18n_prepare(fields)
        async with make_session() as session:
            obj = cls.model(**fields)  # noqa
            session.add(obj)
            await session.commit()
            await session.refresh(obj)
            fields.update({'id': obj.id})
            cls.i18n_release(fields)
            return fields

    @staticmethod
    def row_to_dict(row) -> Dict:
        """Row to dict."""
        row = row.__dict__
        row.pop('_sa_instance_state', None)
        return row

    @classmethod
    async def get_by_id(cls, session, id: int) -> Dict:
        """Get by id."""
        query = await session.execute(select(cls.model).filter_by(id=id))

        try:
            return cls.row_to_dict(query.scalar())
        except Exception as exc:
            print(exc)
            return {}

    @classmethod
    async def update(cls, id: int, fields: Dict, lang: str = None) -> Dict:
        """Update entity."""
        cls.i18n_prepare(fields)
        async with make_session() as session:
            origin = await cls.get_by_id(session, id)

            if not origin:
                return

            for new_name, new_value in fields.items():
                if new_name in origin:
                    if not cls.i18n_fields:
                        origin[new_name] = new_value
                    elif new_name not in cls.i18n_fields:
                        origin[new_name] = new_value
                    else:
                        _src = dict(origin[new_name])
                        _src.update(new_value)
                        origin[new_name] = _src

            origin.pop('id', None)
            await session.execute(
                update(cls.model).filter_by(id=id).values(**origin),
            )
            await session.commit()
            return await cls.read(id, lang=lang)

    @classmethod
    async def delete(cls, id: int) -> Dict:
        """Delete entity."""
        async with make_session() as session:
            await session.execute(delete(cls.model).filter_by(id=id))
            await session.commit()

    @classmethod
    async def translate(cls, id: int, lang: str, fields: Dict):
        """Translate entity."""
        dest_lang = str(lang).replace(' ', '').lower()
        async with make_session() as session:
            raw = await cls.get_by_id(session, id)

            for _f, _v in raw.items():
                if _f in fields and _f in raw and _f in cls.i18n_fields:
                    _v[dest_lang] = str(fields[_f])

            await cls._update_from_dict(session, id, raw)

    @classmethod
    async def _update_from_dict(cls, session, id: int, fields: Dict) -> None:
        """Update entity from dict."""
        query = update(cls.model).filter_by(id=id).values(**fields)
        await session.execute(query)
        await session.commit()
